-- See the Copyright notice at the end of this file.
--
class RECYCLING_POOL[R_ -> RECYCLABLE]
	--
	-- RECYCLABLE objects container.
	--
	-- Useful to implement recycling support in libraries.
	--
	-- Has no sense for expanded types (it will not even compile).
	--
	-- This class is designed to work with and without the garbage collector. It takes advantage of the GC if
	-- it is used, but it also minimises the memory footprint when the GC is ''not'' used.
	--
	-- See also STRING_RECYCLING_POOL, especially tuned for STRING usage.
	--

insert
	MEMORY
		export
			-- to be able to put RECYCLING_POOL objects in a DICTIONARY:
			{ANY} is_expanded_type, is_equal, out_in_tagged_out_memory, is_basic_expanded_type, same_dynamic_type;
			{RECYCLING_POOL} all
		end

creation {ANY}
	make

feature {ANY}
	recycle (a_item: like item) is
			--
			-- Stores the object as being reuseable. Automatically calls the "recycle" feature of the object.
			--
			-- Two notes:
			--
			-- * You should not directly reuse the object after it is recycled; call `item' instead to obtain a
			-- fresh reference. Using it directly will lead to classic problems of double referencing. In a
			-- nutshell, '''be sure not to hold any reference to a recycled object'''.
			--
			-- * If the GC is used it may free some recycled objects so using a RECYCLING_POOL does not interfere
			-- with memory conservation.
			--
			-- See also `item'
			--
		require
			a_item /= Void
		local
			i: INTEGER; done: BOOLEAN; wr: WEAK_REFERENCE[R_]
		do
			a_item.recycle
			if cursor /= no_item then
				from
					i := cursor
				until
					done or else i > ring.upper
				loop
					wr := ring.item(i)
					if wr.item = Void then
						wr.set_item(a_item)
						done := True
						cursor := i
					end
					i := i + 1
				end
				if not done then
					from
						i := ring.lower
					until
						done or else i = cursor
					loop
						wr := ring.item(i)
						if wr.item = Void then
							wr.set_item(a_item)
							done := True
							cursor := i
						end
						i := i + 1
					end
				end
			end
			if not done then
				create wr.set_item(a_item)
				ring.add_last(wr)
				cursor := ring.upper
			end
		ensure
			cursor /= no_item
			ring.valid_index(cursor)
			count = old count + 1
		end

	item: R_ is
			--
			-- Returns a recycled object, if there is one to be obtained. Returns Void otherwise.
			--
			-- See also `recycle'
			--
		local
			i: INTEGER; gc: BOOLEAN
		do
			if cursor /= no_item then
				gc := collecting
				collection_off
				i := next(cursor)
				if i = no_item then
					cursor := no_item
				else
					Result := item_at(i)
					check
						Result /= Void
						ring.item(i).item = Void
					end
				end
				if gc then
					collection_on
				end
			end
		ensure
			Result = Void implies cursor = no_item
			old count > 0 implies count = old count - 1
		end

feature {STRING_RECYCLING_POOL} -- internals
	no_item: INTEGER is -1
			-- Default cursor value when no recycled item is available

	cursor: INTEGER
			-- Optimization: keep the position of the latest saved or retrieved item.

	next (start: INTEGER): INTEGER is
			-- Try to find a recycled object. Returns `no_item' if such an object cannot be found.
		require
			not collecting
			start /= no_item
		local
			i: INTEGER
		do
			Result := no_item
			from
				i := start
			until
				Result /= no_item or else i > ring.upper
			loop
				if ring.item(i).item /= Void then
					Result := i
				end
				i := i + 1
			end
			if Result = no_item then
				from
					i := ring.lower
				until
					Result /= no_item or else i = cursor
				loop
					if ring.item(i).item /= Void then
						Result := i
					end
					i := i + 1
				end
			end
		ensure
			Result /= no_item implies ring.item(Result).item /= Void
		end

	item_at (i: INTEGER): R_ is
			-- Returns the item at the given position.
		require
			not collecting
			i /= no_item
		local
			wr: WEAK_REFERENCE[R_]
		do
			wr := ring.item(i)
			Result := wr.item
			if Result /= Void then
				wr.set_item(Void)
				cursor := i + 1
			else
				cursor := no_item
			end
		end

feature {}
	ring: RING_ARRAY[WEAK_REFERENCE[R_]]
			-- The recycled items

	count: INTEGER is
			-- The number of recycled items still available (may be less than the number of recycle dobjects if
			-- the GC is in use). Used only in assertions.
		local
			i: INTEGER
		do
			from
				i := ring.lower
			until
				i > ring.upper
			loop
				if ring.item(i).item /= Void then
					Result := Result + 1
				end
				i := i + 1
			end
		end

	make is
			-- Create a new recyclable objects container.
		do
			create ring.with_capacity(16, 0)
			cursor := no_item
		end

invariant
	ring.lower = 0

end -- class RECYCLING_POOL
--
-- ------------------------------------------------------------------------------------------------------------
-- Copyright notice below. Please read.
--
-- This file is part of the SmartEiffel standard library.
-- Copyright(C) 1994-2002: INRIA - LORIA (INRIA Lorraine) - ESIAL U.H.P.       - University of Nancy 1 - FRANCE
-- Copyright(C) 2003-2006: INRIA - LORIA (INRIA Lorraine) - I.U.T. Charlemagne - University of Nancy 2 - FRANCE
--
-- Authors: Dominique COLNET, Philippe RIBET, Cyril ADRIAN, Vincent CROIZIER, Frederic MERIZEN
--
-- Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
-- documentation files (the "Software"), to deal in the Software without restriction, including without
-- limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
-- the Software, and to permit persons to whom the Software is furnished to do so, subject to the following
-- conditions:
--
-- The above copyright notice and this permission notice shall be included in all copies or substantial
-- portions of the Software.
--
-- THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
-- LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO
-- EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN
-- AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
-- OR OTHER DEALINGS IN THE SOFTWARE.
--
-- http://SmartEiffel.loria.fr - SmartEiffel@loria.fr
-- ------------------------------------------------------------------------------------------------------------
